/*=====================================================================*/
/*    serrano/prgm/project/bigloo/runtime/Clib/cports.c                */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Thu Jul 23 15:34:53 1992                          */
/*    Last change :  Fri Sep 12 14:39:19 2008 (serrano)                */
/*    -------------------------------------------------------------    */
/*    Input ports handling                                             */
/*=====================================================================*/
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _MSC_VER
#   include <dirent.h>
#   include <sys/time.h>
#else
#   include <io.h>
#   include <windows.h>
#   include <winsock2.h>
#   define pclose _pclose
#   define popen _popen
#   define S_ISDIR( mode ) ((mode & _S_IFDIR) != 0)
#endif
#include <string.h>
#if !defined( sony_news ) && \
    !(defined( NeXT ) && (defined( mc68000 ) || defined( i386 ))) && \
    !defined( _MSC_VER )
#   include <unistd.h>
#endif
#ifndef _MSC_VER
#  include <sys/file.h>
#endif
#include <bigloo.h>
#if( defined( sony_news ) || (defined( NeXT ) && defined( mc68000 )) )
#   include <ctype.h>
#endif

#if POSIX_FILE_OPS
#   include <unistd.h>
#endif

#if HAVE_TERMIOS
#  include <termios.h>
#else 
#  if HAVE_TERMIO
#    include <termio.h>
#  endif
#endif

#if( defined( S_IRUSR ) && \
     defined( S_IWUSR ) && \
     defined( S_IRGRP ) && \
     defined( S_IWGRP ) && \
     defined( S_IROTH ) && \
     defined( S_IWOTH ))
#  define OMOD (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)
#else
#  define OMOD 0666
#endif

#if( !defined( EWOULDBLOCK ) )
#  define EWOULDBLOCK 0
#endif

#if( !defined( EAGAIN ) )
#  define EAGAIN 0
#endif

#if( !defined( EINTR ) )
#  define EINTR 0
#endif

/*---------------------------------------------------------------------*/
/*    Default IO operations                                            */
/*---------------------------------------------------------------------*/
#if POSIX_FILE_OPS
#  define _WRITE write
#  define _LSEEK lseek
#  define _CLOSE close
#  define _FILENO fileno
#  define _FD int
#  define _CREAT( name, mod ) creat( name, mod )
#  define _OPEN( name, flag, mod ) open( name, flag, mod )
#  define _FSTAT fstat
#else
#  define _WRITE posix_write
#  define _LSEEK posix_lseek
#  define _CLOSE posix_close
#  define _FILENO
#  define _FD FILE * 
#  define _CREAT( name, mod ) fopen( name, "wb" )
#  define _OPEN( name, flag, mod ) fopen( name, "a+b" )
#  define _FSTAT( fd, buf ) fstat( fileno( fd ), buf )
#endif

#if( BGL_HAVE_SENDFILE )
extern ssize_t sendfile( int, int, off_t *, size_t );
#  define _sendfile sendfile
#endif

/*---------------------------------------------------------------------*/
/*    compatibility kit ISO C / POSIX 2001.                            */
/*---------------------------------------------------------------------*/
static size_t posix_write( FILE *f, void *buf, size_t count ) {
   return fwrite( buf, 1, count, f );
}

static long posix_lseek( FILE *f, long offset, int whence ) {
   return fseek( f, offset, whence );
}

static int posix_close( FILE *f ) {
   return fclose( f );
}

/*---------------------------------------------------------------------*/
/*    DEBUG_SENDCHARS                                                  */
/*---------------------------------------------------------------------*/
#define DEBUG_SENDCHARS 1
#undef DEBUG_SENDCHARS

/*---------------------------------------------------------------------*/
/*    isascii                                                          */
/*---------------------------------------------------------------------*/
#if( !defined( isascii ) )
#   define isascii( c ) (!((c) & ~0177))
#endif

/*---------------------------------------------------------------------*/
/*    Global variables                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF long default_io_bufsiz;

/*---------------------------------------------------------------------*/
/*    External definitions.                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DECL obj_t bgl_make_input_port( obj_t, FILE *, obj_t, obj_t );
extern obj_t make_string();
extern obj_t make_string_sans_fill();

/*---------------------------------------------------------------------*/
/*    Prototypes                                                       */
/*---------------------------------------------------------------------*/
static bool_t pipe_name_p( char * );
static char *pipe_name(char *);

static size_t strwrite( obj_t, void *, size_t );
static long strseek( void *, long, int );

static size_t procwrite( obj_t, void *, size_t );
static obj_t procflush( obj_t );

/*---------------------------------------------------------------------*/
/*    standard ports ...                                               */
/*---------------------------------------------------------------------*/
static obj_t _stdin, _stdout, _stderr;

/*---------------------------------------------------------------------*/
/*    struct sendfile_info_t ...                                       */
/*    -------------------------------------------------------------    */
/*    This structure is used for the Boehm's collection version        */
/*    7 and higher. The collector require a "blocking" function        */
/*    to be wrapping sendfile invokation. This function receives       */
/*    a sendfile_info_t as parameter and fills the "res" field.        */
/*---------------------------------------------------------------------*/
struct sendfile_info_t {
   int out;
   int in;
   off_t *offset;
   size_t sz;
   long res;
};

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bglerror ...                                                     */
/*---------------------------------------------------------------------*/
static int
bglerror( int err ) {
   switch( err ) {
      case EPIPE: return BGL_IO_SIGPIPE_ERROR;

#if( defined( EBADF ) )
      case EBADF: return BGL_IO_PORT_ERROR;
#endif	 
#if( defined( EBADFD ) )
      case EBADFD: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENAMETOOLONG ) )
      case ENAMETOOLONG: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENFILE ) )
      case ENFILE: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENODEV ) )
      case ENODEV: return BGL_IO_PORT_ERROR;
#endif

#if( defined( ENOMEM ) )
      case ENOMEM: return BGL_IO_WRITE_ERROR;
#endif	 
#if( defined( ENOSPC ) )
      case ENOSPC: return BGL_IO_WRITE_ERROR;
#endif	 
	 
      default: return BGL_IO_ERROR;
   }
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    bgl_feof ...                                                     */
/*    -------------------------------------------------------------    */
/*    We explicitly define this function because on some system        */
/*    feof is a macro.                                                 */
/*---------------------------------------------------------------------*/
int
bgl_feof( obj_t obj ) {
   return feof( (FILE *)PORT_STREAM( obj ) );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    proc_eof ...                                                     */
/*---------------------------------------------------------------------*/
static int
proc_eof( obj_t obj ) {
   return 1;
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    eof_fread ...                                                    */
/*    -------------------------------------------------------------    */
/*    Used, for instance, on input string ports.                       */
/*---------------------------------------------------------------------*/
static long
eof_fread( char *ptr, long size, long nmemb, obj_t port ) {
   return 0;
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    plain_fread ...                                                  */
/*---------------------------------------------------------------------*/
static long
plain_fread( char *ptr, long size, long nmemb, obj_t port ) {
   FILE *stream = PORT_STREAM( port );
   
#ifdef POSIX_FILE_OPS
   long num = size * nmemb;
   long n;

 loop:
   if( (n = read( fileno( stream ), ptr, num ) ) < 0 )
      if( errno == EINTR ) goto loop;

   return n;
#else
   long n;

 loop:
   if( !(n = fread( ptr, size, nmemb, stream )) ) {
      if( ferror( stream ) == EINTR ) {
	 clearerr( stream );
	 goto loop;
      } else {
	 C_SYSTEM_FAILURE( BGL_IO_READ_ERROR, "read", strerror(errno), port );
      }
   }

   return n;
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    console_fread ...                                                */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input, if */
/*    not enough characters are available.                             */
/*    -------------------------------------------------------------    */
/*    This function is prefered to bgl_pipe_fread for consoles because */
/*    it automatically flushes stdout before reading (see getc).       */
/*---------------------------------------------------------------------*/
static long
console_fread( char *ptr, long size, long nmemb, obj_t port ) {
   FILE *stream = PORT_STREAM( port );
   long num = size * nmemb;
   char *buf = ptr;
   int c;

   while( ((c = getc( stream )) != EOF) ) {
      *buf++ = c;

      if( c == '\n' ) break;
      if( --num <= 0 ) break;
   }

   return (long)(buf - ptr);
}

/*---------------------------------------------------------------------*/
/*    long                                                             */
/*    bgl_pipe_fread ...                                               */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input if  */
/*    insufficient characters are available.                           */
/*---------------------------------------------------------------------*/
long
bgl_pipe_fread( char *ptr, long size, long nmemb, obj_t port ) {
   FILE *stream = PORT_STREAM( port );
   int fd = fileno( stream );
   
#ifdef POSIX_FILE_OPS
   long num = size * nmemb;
   long n;

 loop:
   if( (n = read( fd, ptr, num )) < 0 ) {
      if( errno == EINTR ) goto loop;
      C_SYSTEM_FAILURE( bglerror( errno ), "read", strerror( errno ), port );
   }

#if( BGL_HAVE_SELECT )
   if( n == 0 ) {
      /* is it a true end-of-file or a simple lack of input? */
      struct timeval timeout;
      fd_set readfds;
      
      timeout.tv_sec = 0;
      timeout.tv_usec = 10000;

      FD_ZERO( &readfds );
      FD_SET( fd, &readfds );

      if( (n = select( fd + 1, &readfds, NULL, NULL, &timeout ) <= 0 ) ) {
	 if( n == 0 ) {
	    /* FIXME */
	    fprintf( stderr, "%s:%d:LOOPING on select timeout\n", __FILE__, __LINE__ );
	    goto loop;
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			      "read/timeout",
			      strerror( errno ),
			      port );
	 }
      }
   }
#endif
   return n;
#else
   return console_fread( ptr, size, nmemb, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    proc_read ...                                                    */
/*---------------------------------------------------------------------*/
static long
proc_read( char *b, long s, long n, obj_t port ) {
   long l = s * n;
   obj_t buf = CREF(port)->input_port_t.pbuffer;

 loop:

   if( STRINGP( buf ) ) {
      /* won't read because the proc buffer is already filled */
      char *s = BSTRING_TO_STRING( buf );
      long p = CREF( port )->input_port_t.pbufpos;
      long r = STRING_LENGTH( buf ) - p;

      if( r <= l ) {
	 memmove( b, &s[ p ], r );
	 CREF( port )->input_port_t.pbuffer = BFALSE;
	 CREF( port )->input_port_t.pbufpos = 0;
	 return r;
      } else {
	 memmove( b, &s[ p ], l );
	 CREF( port )->input_port_t.pbufpos += l;
	 return l;
      }
   } else {
      /* invoke the procedure to fill the buffer */
      obj_t proc = CREF( port )->input_port_t.port.name;
      obj_t nbuf = PROCEDURE_ENTRY( proc )( proc, BEOA );

      if( STRINGP( nbuf ) ) {
	 buf = CREF( port )->input_port_t.pbuffer = nbuf;
	 goto loop;
      } else {
	 if( nbuf == BFALSE ) {
	    /* eof has been reached */
	    CREF( port )->input_port_t.eof = 1;
	    return 0;
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "input-procedure-port",
			      "Procedure result must be #f, or a string",
			      nbuf);
	 }
      }
   }
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    timeout_set_port_blocking ...                                    */
/*---------------------------------------------------------------------*/
#if( BGL_HAVE_FCNTL )
static void
timeout_set_port_blocking( char *fun, int fd, int bool ) {
   int val;

   if( (val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_ERROR, fun, strerror( errno ), BINT( fd ) );
   }

   if( !bool ) 
      val |= O_NONBLOCK;
   else 
      val &= ~O_NONBLOCK;

   if( fcntl( fd, F_SETFL, val ) < 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_ERROR, fun, strerror( errno ), BINT( fd ) );
   }
}
#endif 

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    sysread_with_timeout ...                                         */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input if  */
/*    insufficient characters are available.                           */
/*---------------------------------------------------------------------*/
static long
sysread_with_timeout( char *ptr, long size, long nmemb, obj_t port ) {
#if( defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   int fd = fileno( (FILE *)PORT_STREAM( port ) );
   struct bgl_input_timeout *tmt = PORT( port ).timeout;
   long (*sysread)() = tmt->sysread;
   long n;

   long to = tmt->timeout;

   fd_set readfds;
   struct timeval timeout;

   FD_ZERO( &readfds );
   FD_SET( fd, &readfds );

   timeout.tv_sec = to / 1000000;
   timeout.tv_usec = to % 1000000;

loop:   
   if( (n = select( fd + 1, &readfds, NULL, NULL, &timeout )) <= 0 ) {
      if( n == 0 ) {
	 C_SYSTEM_FAILURE( BGL_IO_TIMEOUT_ERROR,
			   "read/timeout",
			   "time limit exceeded",
			   port );
      } else {
	 /* MS: 23 Jan 2008. No attention was paid to EINTR */
	 /* I'm not 100% sure that this new test is correct */
	 if( errno == EINTR ) {
	    fprintf( stderr, "*** DEBUG SELECT receives EINTR: %s:%d\n",
		     __FILE__, __LINE__ );
	    FD_ZERO( &readfds );
	    FD_SET( fd, &readfds );

	    timeout.tv_sec = to / 1000000;
	    timeout.tv_usec = to % 1000000;
	    goto loop;
	 }
	 
	 C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			   "read/timeout",
			   strerror( errno ),
			   port );
      }
   } else {
      n = sysread( ptr, size, nmemb, port );

      return n;
   }
#else
   return INPUT_PORT( port ).sysread( ptr, size, nmemb, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    strseek ...                                                      */
/*---------------------------------------------------------------------*/
static long
strseek( void *port, long offset, int whence ) {
   obj_t buf = OUTPUT_PORT( port ).buf;
   int len = STRING_LENGTH( buf );
   int cnt = OUTPUT_PORT( port ).cnt;

   if( (offset < 0) || (offset >= (len - cnt)) )
      return 1;

   OUTPUT_PORT( port ).ptr = &STRING_REF( buf, offset );
   OUTPUT_PORT( port ).cnt = len - offset;

   return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_output_flush ...                                             */
/*    -------------------------------------------------------------    */
/*    There is no room for str and cnt has not been decremented.       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_output_flush( obj_t port, char *str, size_t slen ) {
   if( PORT( port ).kindof == KINDOF_CLOSED ) {
      return BFALSE;
   } else {
      obj_t buf = OUTPUT_PORT( port ).buf;
      int len = STRING_LENGTH( buf );
      size_t (*syswrite)( void *, void *, size_t ) =
	 OUTPUT_PORT( port ).syswrite;
      char *base = BSTRING_TO_STRING( buf );
      size_t cnt = OUTPUT_PORT( port ).cnt;
      long use = len - cnt;
      int n;

#if DEBUG   
      fprintf( stderr, "%s:%d>>> bgl_output_flush port=%p slen=%d bufmode=%d cnt=%d\n",
	       __FILE__, __LINE__, port, slen, OUTPUT_PORT( port ).bufmode,
	       OUTPUT_PORT( port ).cnt );
#endif   
      
      /* copy the buffer, if needed */
      if( OUTPUT_PORT( port ).bufmode != BGL_IOEBF ) {
	 /* some ports use extensible buffers, e.g.   */
	 /* string ports which must not been written. */
	 while( use > 0 ) {
	    if( (n = syswrite( PORT_STREAM( port ), base, use )) < 0 ) {
	       if( errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR ) {
		  continue;
	       } else {
		  C_SYSTEM_FAILURE( bglerror( errno ),
				    "write/display",
				    strerror( errno ),
				    port );
	       }
	    } else {
	       use -= n;
	       base += n;
	    }
	 }
   
	 OUTPUT_PORT( port ).ptr = &STRING_REF( buf, 0 );
	 OUTPUT_PORT( port ).cnt = STRING_LENGTH( buf );
      
	 /* copy the string, always */
	 while( slen > 0 ) {
	    if( (n = syswrite( PORT_STREAM( port ), str, slen )) < 0 ) {
	       if( errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR ) {
		  continue;
	       } else {
		  C_SYSTEM_FAILURE( bglerror( errno ),
				    "write/display",
				    strerror( errno ),
				    port );
	       }
	    } else {
	       slen -= n;
	       str += n;
	    }
	 } 
      } else {
	 /* this is an exensible buffer, that we increase iff it is full */
	 if( (slen > 0) || (cnt == 0) ) {
	    if( (n = syswrite( PORT_STREAM( port ), str, slen )) < 0 ) {
	       C_SYSTEM_FAILURE( bglerror( errno ),
				 "write/display",
				 strerror( errno ),
				 port );
	    }
	 }
      }


#if DEBUG   
      fprintf( stderr, "%s:%d<<< bgl_output_flush port=%p cnt=%d\n",
	       __FILE__, __LINE__, port, OUTPUT_PORT( port ).cnt );
#endif   
      return port;
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_write ...                                                    */
/*---------------------------------------------------------------------*/
obj_t
bgl_write( obj_t op, char *str, size_t sz ) {
#if DEBUG   
   fprintf( stderr, "%s:%d >>> bgl_write op=%p len=%d cnt=%d mode=%d [%s:%d]\n",
	    __FILE__, __LINE__,
	    op, STRING_LENGTH( OUTPUT_PORT( op ).buf ),
	    OUTPUT_PORT( op ).cnt, OUTPUT_PORT( op ).bufmode, str, sz );
#endif
   if( OUTPUT_PORT( op ).cnt > sz ) {
      if( OUTPUT_PORT( op ).bufmode == BGL_IOLBF ) {
	 while( sz-- > 0 ) {
	    char c = *str++;
	    if( c == '\n' ) {
	       OUTPUT_PORT( op ).cnt--;
	       *OUTPUT_PORT( op ).ptr++ = c;
	       bgl_output_flush( op, 0, 0 );
	    } else {
	       *OUTPUT_PORT( op ).ptr++ = c;
	       OUTPUT_PORT( op ).cnt--;
	    }
	 }
      } else {
	 memcpy( OUTPUT_PORT( op ).ptr, str, sz );
	 OUTPUT_PORT( op ).ptr += sz;
	 OUTPUT_PORT( op ).cnt -= sz;
      }
      
#if DEBUG   
   fprintf( stderr, "%s:%d <<< bgl_write op=%p len=%d cnt=%d mode=%d [%s:%d]\n",
	    __FILE__, __LINE__,
	    op, STRING_LENGTH( OUTPUT_PORT( op ).buf ),
	    OUTPUT_PORT( op ).cnt, OUTPUT_PORT( op ).bufmode, str, sz );
#endif
      return op;
   } else {
      return bgl_output_flush( op, str, sz );
   }
}

/*---------------------------------------------------------------------*/
/*     bgl_make_output_port ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_make_output_port( obj_t name,
		      void *stream,
		      obj_t kindof,
		      obj_t buf,
		      size_t (*write)(),
		      long (*seek)(),
		      int (*close)() ) {
   obj_t new_output_port;

   if( !STRINGP( buf ) ) {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"make-output-port",
			"Illegal buffer",
			buf );
   }
   
   new_output_port = GC_MALLOC( OUTPUT_PORT_SIZE );
   
   new_output_port->output_port_t.port.header =
      MAKE_HEADER( OUTPUT_PORT_TYPE, 0 );
   
   new_output_port->port_t.name = name;
   new_output_port->port_t.stream = stream;
   new_output_port->port_t.kindof = kindof;
   new_output_port->port_t.chook = BUNSPEC;
   new_output_port->port_t.userdata = BUNSPEC;
   new_output_port->port_t.timeout = 0L;
   new_output_port->port_t.sysseek = seek;
   new_output_port->port_t.sysclose = close;
   new_output_port->output_port_t.syswrite = write;
   new_output_port->output_port_t.sysflush = 0L;

   new_output_port->output_port_t.bufmode = BGL_IOFBF;
   bgl_output_port_buffer_set( BREF( new_output_port ), buf );
   
   return BREF( new_output_port );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_output_port_buffer_set ...                                   */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_output_port_buffer_set( obj_t port, obj_t buf ) {
   if( !STRINGP( buf ) ) {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"output-port-buffer-set!",
			"Illegal buffer",
			buf );
   }
   
   OUTPUT_PORT( port ).buf = buf;
   OUTPUT_PORT( port ).ptr = &STRING_REF( buf, 0 );
   OUTPUT_PORT( port ).cnt = STRING_LENGTH( buf );
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    bgl_output_port_timeout_set ...                                  */
/*---------------------------------------------------------------------*/
bool_t
bgl_output_port_timeout_set( obj_t port, long timeout ) {
#if defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL
   if( PORT_IS_OS( port ) && (timeout >= 0) ) {
      if( timeout == 0 ) {
	 if( PORT( port ).timeout ) {
	    timeout_set_port_blocking( "output-port-timeout-set!",
				       (int)PORT_STREAM( port ),
				       1 );
	    PORT( port ).timeout = 0;
	 }
      } else {
	 if( !PORT( port ).timeout ) {
	    timeout_set_port_blocking( "output-port-timeout-set!",
				       (int)PORT_STREAM( port ),
				       0 );
	 }
	 PORT( port ).timeout = (void *)timeout;
      }

      return 1;
   }
#endif

   return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_file_to_output_port ...                                      */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_file_to_output_port( FILE *f, obj_t buf ) {
   return bgl_make_output_port( string_to_bstring( "<c-port>" ),
				(void *)_FILENO( f ),
				KINDOF_FILE, 
				buf, 
				(size_t (*)())_WRITE,
				(long (*)())_LSEEK,
				_CLOSE );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_output_file ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_output_file( obj_t name, obj_t buf ) {
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      FILE *file = popen( pipe_name( cname ), "w" );
      
      if( !file )
         return BFALSE;

      setvbuf( file, 0, _IONBF, 0 );
	       
      return bgl_make_output_port( name, file,
				   KINDOF_PIPE,
				   buf,
				   posix_write,
				   posix_lseek,
				   pclose );
   } else
#endif
   {
      _FD fd;
      
      if( strcmp( cname, "null:" ) == 0 ) {
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif
      }

      if( !(fd = _CREAT( cname, OMOD )) )
         return BFALSE;
      else
	 return bgl_make_output_port( name, (void *)fd,
				      KINDOF_FILE,
				      buf,
				      (size_t (*)())_WRITE,
				      (long (*)())_LSEEK,
				      _CLOSE );
   }
}

/*---------------------------------------------------------------------*/
/*    bgl_append_output_file ...                                       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_append_output_file( obj_t name, obj_t buf ) {
   _FD fd;
   
   if( !(fd = _OPEN( BSTRING_TO_STRING( name ), O_CREAT | O_WRONLY, OMOD )) )
      return BFALSE;
   else {
      if( _LSEEK( fd, 0, SEEK_END ) < 0 ) {
	 _CLOSE( fd );
	 return BFALSE;
      } else {
	 return bgl_make_output_port( name, (void *)fd,
				      KINDOF_FILE,
				      buf, 
				      (size_t (*)())_WRITE,
				      (long (*)())_LSEEK,
				      _CLOSE );
      }
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_output_string ...                                       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_output_string( obj_t buf ) {
   obj_t port = bgl_make_output_port( string_to_bstring( "string" ), 0L,
				      KINDOF_STRING,
				      buf,
				      strwrite, strseek, 0 );
   PORT_STREAM( port ) = port;
   OUTPUT_PORT( port ).bufmode = BGL_IOEBF;
   OUTPUT_PORT( port ).sysflush = get_output_string;

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_output_procedure ...                                    */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_output_procedure( obj_t proc, obj_t flush, obj_t buf ) {
   obj_t port = bgl_make_output_port( string_to_bstring( "procedure" ), 0L,
				      KINDOF_PROCEDURE,
				      make_string_sans_fill( 0 ),
				      procwrite, 0L, 0L );
   obj_t udata = create_vector( 4 );
   PORT_STREAM( port ) = port;
   OUTPUT_PORT( port ).bufmode = BGL_IONB;
   OUTPUT_PORT( port ).sysflush = procflush;
   PORT( port ).userdata = udata;
   VECTOR_SET( udata, 0, proc );
   VECTOR_SET( udata, 1, buf );
   VECTOR_SET( udata, 2, flush );

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    get_output_string ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
get_output_string( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_STRING ) {
      obj_t buf = OUTPUT_PORT( port ).buf;
      int cnt = OUTPUT_PORT( port ).cnt;
	 
      return string_to_bstring_len( BSTRING_TO_STRING( buf ),
				    STRING_LENGTH( buf ) - cnt );
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"get-output-string",
			"Not a string port",
			port );
      return BUNSPEC;
   }
}

/*---------------------------------------------------------------------*/
/*    bgl_close_output_port ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_close_output_port( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_CLOSED )
      return port;
      
   if( (port == _stdout) || (port == _stderr) ) {
      bgl_output_flush( port, 0, 0 );
      return port;
   } else {
      obj_t res = port;
      obj_t chook = PORT_CHOOK( port );
   
      if( PORT( port ).kindof == KINDOF_STRING ) {
	 obj_t buf = OUTPUT_PORT( port ).buf;
	 int cnt = OUTPUT_PORT( port ).cnt;

#if DEBUG	 
	 fprintf( stderr, "%s:%d, close: port=%p length=%d cnt=%d\n",
		  __FILE__, __LINE__,
		  port,
		  STRING_LENGTH( buf ) - cnt, cnt );
#endif	 
	 res = bgl_string_shrink( buf, STRING_LENGTH( buf ) - cnt );
      } else {
	 bgl_output_flush( port, 0, 0 );
      }
      
      PORT( port ).kindof = KINDOF_CLOSED;
      
      if( PORT( port ).sysclose )
	 PORT( port ).sysclose( PORT_STREAM( port ) );

      if( PROCEDUREP( chook ) ) {
	 if( PROCEDURE_ARITY( chook ) == 1 ) {
	    PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "close-output-port",
			      "illegal close hook arity",
			      chook );
	 }
      }

      OUTPUT_PORT( port ).buf = BFALSE;
      return res;
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_make_input_port ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_make_input_port( obj_t name, FILE *file, obj_t kindof, obj_t buf ) {
   /* An input port cannot be allocated as an atomic object    */
   /* because it holds a buffer and a name that are GC objects */
   obj_t new_input_port = GC_MALLOC( INPUT_PORT_SIZE );

   new_input_port->port_t.header = MAKE_HEADER( INPUT_PORT_TYPE, 0 );
   new_input_port->port_t.kindof = kindof;
   new_input_port->port_t.name = name;
   new_input_port->port_t.stream = file;
   new_input_port->port_t.timeout = 0L;
   new_input_port->port_t.sysseek = 0L;
   new_input_port->port_t.chook = BUNSPEC;
   new_input_port->port_t.userdata = BUNSPEC;
   new_input_port->input_port_t.filepos = 0;
   new_input_port->input_port_t.fillbarrier = -1;
   new_input_port->input_port_t.eof = 0;
   new_input_port->input_port_t.matchstart = 0;
   new_input_port->input_port_t.matchstop = 0;
   new_input_port->input_port_t.forward = 0;
   new_input_port->input_port_t.bufpos = 1;
   new_input_port->input_port_t.lastchar = '\n';
   new_input_port->input_port_t.syseof = bgl_feof;
   new_input_port->input_port_t.pbuffer = BFALSE;
   new_input_port->input_port_t.pbufpos = 0;
   new_input_port->input_port_t.gzip = BFALSE;
   new_input_port->input_port_t.buf = buf;

   switch( (long) kindof ) {
      case (long)KINDOF_CONSOLE:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.sysread = console_fread;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
	 break;

#if( HAVE_PIPE )
      case (long)KINDOF_PIPE:
	 new_input_port->port_t.sysclose = pclose;
	 new_input_port->input_port_t.sysread = bgl_pipe_fread;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
	 break;

      case (long)KINDOF_PROCPIPE:
#endif
      case (long)KINDOF_SOCKET:
	 new_input_port->port_t.sysclose = fclose;
	 new_input_port->input_port_t.sysread = bgl_pipe_fread;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
	 break;

      case (long)KINDOF_FILE:
	 new_input_port->port_t.sysclose = fclose;
	 new_input_port->input_port_t.sysread = plain_fread;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
	 break;

      case (long)KINDOF_PROCEDURE:
      case (long)KINDOF_GZIP:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.syseof = proc_eof;
	 new_input_port->input_port_t.sysread = proc_read;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
	 break;

      case (long)KINDOF_STRING:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.sysread = eof_fread;
	 break;
	 
      default:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.sysread = plain_fread;
	 STRING_SET( new_input_port->input_port_t.buf, 0, '\0' );
   }


   return BREF( new_input_port );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_input_port_buffer_set ...                                    */
/*---------------------------------------------------------------------*/
void
bgl_input_port_buffer_set( obj_t ip, obj_t buffer ) {
   BGL_INPUT_PORT_BUFFER( ip ) = buffer;
      
   INPUT_PORT( ip ).matchstart = 0;
   INPUT_PORT( ip ).matchstop = 0;
   INPUT_PORT( ip ).forward = 0;
   INPUT_PORT( ip ).bufpos = 1;
   INPUT_PORT( ip ).lastchar = '\n';

   if( PORT( ip ).kindof != KINDOF_STRING ) {
      STRING_SET( buffer, 0 , '\0' );
   }
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    bgl_input_port_timeout_set ...                                   */
/*---------------------------------------------------------------------*/
bool_t
bgl_input_port_timeout_set( obj_t port, long timeout ) {
#if defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL
   if( (timeout >= 0) &&
       ((PORT(port).kindof == KINDOF_FILE) ||
	(PORT(port).kindof == KINDOF_PIPE) ||
	(PORT(port).kindof == KINDOF_PROCPIPE) ||
	(PORT(port).kindof == KINDOF_CONSOLE) ||
	(PORT(port).kindof == KINDOF_SOCKET)) ) {
      if( timeout == 0 ) {
	 struct bgl_input_timeout *to = PORT( port ).timeout;

	 if( to ) {
	    INPUT_PORT( port ).sysread = to->sysread;

	    timeout_set_port_blocking( "input-port-timeout-set!",
				       fileno( (FILE *)PORT_STREAM( port ) ),
				       1 );
	 }

	 return 0;
      } else {
	 if( PORT( port ).timeout ) {
	    ((struct bgl_input_timeout *)PORT( port ).timeout)->timeout = timeout;
	 } else {
	    struct bgl_input_timeout *to =
	       (struct bgl_input_timeout *)
	       GC_MALLOC( sizeof( struct bgl_input_timeout ) );

	    to->timeout = timeout;
	    to->sysread = INPUT_PORT( port ).sysread;

	    if( (int)PORT_STREAM( port ) == -1 ) {
	       C_SYSTEM_FAILURE( bglerror( errno ),
				 "input-port-timeout-set!",
				 "Illegal input-port",
				 port );
	    }

	    PORT( port ).timeout = (FILE *)to;
	 }

	 INPUT_PORT( port ).sysread = &sysread_with_timeout;
	 timeout_set_port_blocking( "input-port-timeout-set!",
				    fileno( (FILE *)PORT_STREAM( port ) ),
				    0);
	 return 1;
      }
   }
#endif

   return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_input_pipe ...                                          */
/*---------------------------------------------------------------------*/
obj_t
bgl_open_input_pipe( obj_t name, obj_t buffer ) {
#if( HAVE_PIPE )
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

   if( !(file = popen( cname, "r" )) )
      return BFALSE;

   /* we use our own buffer */
   setvbuf( file, NULL, _IONBF, 0 );  

   return bgl_make_input_port( name, file, KINDOF_PIPE, buffer );
#else
   return BFALSE;
#endif
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_input_resource ...                                      */
/*    -------------------------------------------------------------    */
/*    Resources are only meaning in Java.                              */
/*---------------------------------------------------------------------*/
obj_t
bgl_open_input_resource( obj_t name, obj_t buffer ) {
   return BFALSE;
}

/*---------------------------------------------------------------------*/
/*    bgl_open_input_file ...                                          */
/*    -------------------------------------------------------------    */
/*    !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!!  */
/*    We fill up its associated buffer when opening an input port.     */
/*    -------------------------------------------------------------    */
/*    This function open two kind of files. Regular file and Unix      */
/*    like pipes when the file name is something like "| ...".         */
/*---------------------------------------------------------------------*/
obj_t
bgl_open_input_file( obj_t name, obj_t buffer ) {
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      if( !(file = popen( pipe_name( cname ), "r" ) ) )
	 return BFALSE;

      /* we use our own buffer */
      setvbuf( file, NULL, _IONBF, 0 );  

      return bgl_make_input_port( name, file, KINDOF_PIPE, buffer );
   } else
#endif
   {
      if (strcmp( cname, "null:" ) == 0)
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif

       if( !(file = fopen( cname, "rb" )) ) {
	  return BFALSE;
       } else {
	  /* we use our own buffer */
	  setvbuf( file, NULL, _IONBF, 0 );  

	  return bgl_make_input_port( name, file, KINDOF_FILE, buffer );
       }
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_file_to_buffered_input_port ...                              */
/*---------------------------------------------------------------------*/
obj_t
bgl_file_to_buffered_input_port( obj_t name, FILE *file, obj_t buffer ) {
   if( file == stdin )
      return bgl_make_input_port( name, stdin, KINDOF_CONSOLE, buffer );
   else {
      return bgl_make_input_port( name, file, KINDOF_FILE, buffer );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_file_to_input_port ...                                       */
/*---------------------------------------------------------------------*/
obj_t
bgl_file_to_input_port( FILE *file ) {
   obj_t buffer = make_string_sans_fill( default_io_bufsiz );
   
   return bgl_file_to_buffered_input_port( string_to_bstring( "file" ),
					   file,
					   buffer );
}

/*---------------------------------------------------------------------*/
/*    bgl_open_input_string ...                                        */
/*    -------------------------------------------------------------    */
/*    The start index is in bounds.                                    */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_input_string( obj_t string, int start ) {
   obj_t port;
   long bufsiz = STRING_LENGTH( string ) - start;
   obj_t buffer = make_string_sans_fill( bufsiz );

   
   memcpy( &STRING_REF( buffer, 0 ), &(STRING_REF( string, start )), bufsiz );
   port = bgl_make_input_port( string_to_bstring( "[string]" ),
			       0L,
			       KINDOF_STRING,
			       buffer );

   CREF(port)->input_port_t.eof = 1;
   CREF(port)->input_port_t.bufpos = bufsiz + 1;

   return port;
}

/*---------------------------------------------------------------------*/
/*    bgl_open_input_procedure ...                                     */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_input_procedure( obj_t fun, obj_t buffer ) {
   if( PROCEDURE_CORRECT_ARITYP( fun, 0 ) ) {
      obj_t port = bgl_make_input_port( string_to_bstring( "[procedure]" ),
					0L,
					KINDOF_PROCEDURE,
					buffer );

      CREF( port )->port_t.stream = port;
      CREF( port )->port_t.name = fun;
      CREF( port )->input_port_t.pbuffer = BUNSPEC;
      CREF( port )->input_port_t.pbufpos = 0;

      return port;
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"open-input-procedure",
			"Illegal procedure arity",
			fun );
   }
}

/*---------------------------------------------------------------------*/
/*    bgl_open_input_gzip_port ...                                     */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_input_gzip_port( obj_t fun, obj_t in, obj_t buffer ) {
   if( PROCEDURE_CORRECT_ARITYP( fun, 0 ) ) {
      obj_t port = bgl_make_input_port( PORT( in ).name,
					0L,
					KINDOF_GZIP,
					buffer );

      CREF( port )->port_t.stream = port;
      CREF( port )->port_t.name = fun;
      CREF( port )->input_port_t.pbuffer = BUNSPEC;
      CREF( port )->input_port_t.pbufpos = 0;
      CREF( port )->input_port_t.gzip = in;

      return port;
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"open-input-gzip-port",
			"Illegal procedure arity",
			fun );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_open_input_c_string ...                                      */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_input_c_string( char *c_string ) {
   obj_t port;
   long bufsiz = (long)strlen( c_string );
   obj_t buffer = string_to_bstring_len( c_string, bufsiz );
   char *new_string = 0;

   port = bgl_make_input_port( string_to_bstring( "[c_string]" ),
			       0L,
			       KINDOF_STRING,
			       buffer );

   CREF(port)->input_port_t.eof = 1;
   /* MS 25 Aug 2008, used to be bufpos = bufsiz + 1 */
   CREF(port)->input_port_t.bufpos = bufsiz;

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_reopen_input_c_string ...                                    */
/*    -------------------------------------------------------------    */
/*    Simply changes the input buffer of an input string. Does not     */
/*    re-allocate a brand new input-port.                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_reopen_input_c_string( obj_t port, char *c_string ) {
   long bufsiz = (long)strlen( c_string );

   if( BGL_INPUT_PORT_BUFSIZ( port ) < (bufsiz + 1) ) {
      CREF(port)->input_port_t.buf = make_string_sans_fill( bufsiz + 1 );
   }

   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   CREF(port)->input_port_t.matchstart = 0;
   CREF(port)->input_port_t.matchstop = 0;
   CREF(port)->input_port_t.forward = 0;
   CREF(port)->input_port_t.lastchar = '\n';
   strcpy( &RGC_BUFFER_REF( port, 0 ), c_string );

   return port;
}

/*---------------------------------------------------------------------*/
/*    bgl_close_input_port ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_close_input_port( obj_t port ) {
   if( INPUT_PORTP( port ) ) {
      if( (PORT( port ).kindof != KINDOF_CLOSED) &&
	  (PORT( port ).kindof != KINDOF_CONSOLE) ) {
	 obj_t chook = PORT_CHOOK( port );

	 if( PORT( port ).sysclose )
	    PORT( port ).sysclose( PORT_STREAM( port ) );
	 
	 INPUT_PORT( port ).eof = 1;
      	 PORT( port ).kindof = KINDOF_CLOSED;
	 PORT( port ).sysclose = 0L;
	 
	 if( PROCEDUREP( chook ) ) {
	    if( PROCEDURE_ARITY( chook ) == 1 ) {
	       PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	    } else {
	       C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
				 "close-input-port",
				 "illegal close hook arity",
				 chook );
	    }
	 }
      }

   }

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_seek ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_input_port_seek( obj_t port, long pos ) {
   if( INPUT_PORT_ON_FILEP( port ) ) {
      if( fseek( PORT_STREAM( port ), pos, SEEK_SET ) )
	 return BFALSE;
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).eof = 0;
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).forward = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).lastchar = '\n';
      RGC_BUFFER_SET( port, 0, '\0' );

      return BTRUE;
   }

   if( INPUT_PORT_ON_STRINGP( port )
       && (pos < BGL_INPUT_PORT_BUFSIZ( port )) ) {
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).matchstart = pos;
      INPUT_PORT( port ).matchstop = pos;
      INPUT_PORT( port ).forward = pos;

      return BTRUE;
   }

   return BFALSE;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_output_port_seek ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_output_port_seek( obj_t port, long pos ) {
   long (*sysseek)() = PORT( port ).sysseek;

   if( sysseek ) {
      return sysseek( PORT_STREAM( port ), pos, SEEK_SET ) ? BFALSE : BTRUE;
   } else {
      return BFALSE;
   }
}
   
/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_reopen ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_input_port_reopen( obj_t port ) {
   FILE *nf;

   if( !INPUT_PORT_ON_FILEP( port ) ) {
      if( INPUT_STRING_PORTP( port ) ) {
 	 return bgl_input_port_seek( port, 0 );
      } else {
	 return BFALSE;
      }
   }
   
   nf = freopen( BSTRING_TO_STRING( PORT( port ).name ),
		 "r",
		 (FILE *)PORT_STREAM( port ) );

   if( !nf ) return BFALSE;

   PORT_STREAM( port ) = nf;
      
   /* we use our own buffer */
   setvbuf( (FILE *)nf, NULL, _IONBF, 0 );

   INPUT_PORT( port ).filepos = 0;
   INPUT_PORT( port ).eof = 0;
   INPUT_PORT( port ).matchstart = 0;
   INPUT_PORT( port ).matchstop = 0;
   INPUT_PORT( port ).forward = 0;
   INPUT_PORT( port ).bufpos = 1;
   INPUT_PORT( port ).lastchar = '\n';
   RGC_BUFFER_SET( port, 0, '\0' );

   return BTRUE;
}

/*---------------------------------------------------------------------*/
/*    obj                                                              */
/*    reset_console ...                                                */
/*    -------------------------------------------------------------    */
/*    We flush input port, for ^C to work correctly within the         */
/*    interpreter. The only place this function is called is in the    */
/*    REPL (see Eval/eval.scm).                                        */
/*---------------------------------------------------------------------*/
obj_t
reset_console( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_CONSOLE ) {
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).lastchar = '\n';
      RGC_BUFFER_SET( port, 0, '\0' );
   }

   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*     bgl_init_io ...                                                 */
/*---------------------------------------------------------------------*/
void
bgl_init_io() {
#if( !defined( _SBFSIZ ) )
#   define _SBFSIZ 1
#endif
   obj_t denv = BGL_CURRENT_DYNAMIC_ENV();
   
   default_io_bufsiz = BUFSIZ * _SBFSIZ;

   _stdout = bgl_make_output_port( string_to_bstring( "stdout" ),
				   (void *)_FILENO( stdout ),
				   KINDOF_CONSOLE,
				   make_string_sans_fill( 512 ),
				   (size_t (*)())_WRITE,
				   (long (*)())_LSEEK,
				   _CLOSE );
   _stderr = bgl_make_output_port( string_to_bstring( "stderr" ),
				   (void *)_FILENO( stderr ),
				   KINDOF_CONSOLE,
				   make_string_sans_fill( 0 ), 
				   (size_t (*)())_WRITE,
				   (long (*)())_LSEEK,
				   _CLOSE );
   _stdin = bgl_make_input_port( string_to_bstring( "stdin" ),
				 stdin,
				 KINDOF_CONSOLE,
				 make_string_sans_fill( default_io_bufsiz ) );

   OUTPUT_PORT( _stdout ).bufmode = BGL_IOLBF;
   
   BGL_ENV_CURRENT_OUTPUT_PORT_SET( denv, _stdout );
   BGL_ENV_CURRENT_ERROR_PORT_SET( denv, _stderr );
   BGL_ENV_CURRENT_INPUT_PORT_SET( denv, _stdin );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_end_io ...                                                   */
/*---------------------------------------------------------------------*/
void
bgl_end_io() {
   bgl_close_output_port( _stdout );
   bgl_close_output_port( _stderr );
}

/*---------------------------------------------------------------------*/
/*    fexists ...                                                      */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
fexists( char *name ) {
#if( HAVE_PIPE )
   if( pipe_name_p( name ) ) {
      return 1;
   }
#else
   if( pipe_name_p(name) ) {
      return 0;
   }
#endif

# ifndef _MSC_VER
    return !access( name, F_OK );
# else
    /* !!!!! verify semantics of Unix' access */
    return !_access( name, 0 );        
# endif
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    reset_eof ...                                                    */
/*    -------------------------------------------------------------    */
/*    The function erase the end-of-file of input console port.        */
/*    This allows, restart reading after a ^D.                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
reset_eof( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_CONSOLE ) {
      /* we forget about EOF */
      INPUT_PORT( port ).eof = 0;

      /* we cleanup buffer   */
      reset_console( port );

      /* we clear errors.    */
      clearerr( stdin );

      return 1;
   }
   else
      return 0;
}

/*---------------------------------------------------------------------*/
/*    static bool_t                                                    */
/*    pipe_name_p ...                                                  */
/*    -------------------------------------------------------------    */
/*    Is a file name a pipe name ? A pipe name start by the            */
/*    sequence "| ".                                                   */
/*---------------------------------------------------------------------*/
static bool_t
pipe_name_p( char *name ) {
   return( (name[ 0 ] == '|') && (name[ 1 ] == ' ') );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    pipe_name ...                                                    */
/*    -------------------------------------------------------------    */
/*    Pipe name to name translation.                                   */
/*---------------------------------------------------------------------*/
static char *
pipe_name( char *pipe_name ) {
   return (pipe_name + 1);
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    directoryp ...                                                   */
/*    -------------------------------------------------------------    */
/*    Is a file a directory?                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
directoryp( char *name ) { 
   struct stat buf;

   if( stat( name, &buf ) == -1 )
      return 0;

   return S_ISDIR( buf.st_mode & S_IFMT );
}

/*---------------------------------------------------------------------*/
/*    REGULAR_FILE_NAMEP ...                                           */
/*---------------------------------------------------------------------*/
#define REGULAR_FILE_NAMEP( f ) \
   ((f[ 0 ] != '.') || ((f[ 1 ] != 0) && ((f[ 1 ] != '.') || f[ 2 ] != 0)))

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    directory_to_list ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
directory_to_list( char *name ) {
   obj_t res = BNIL;
#ifndef _MSC_VER
   DIR *dir;
   struct dirent *dirent;

   if( (dir = opendir( name )) ) {
      while( (dirent = readdir( dir )) ) {
	 char *fname = dirent->d_name;

	 if( REGULAR_FILE_NAMEP( fname ) ) 
	    res = MAKE_PAIR( string_to_bstring( fname ), res );
      }
      closedir( dir );
   }
#else
   char *const path = (char *)malloc( strlen( name ) + 2 + 1 );

   strcpy( path, name );
   strcat( path, "\\*" );

   {
      WIN32_FIND_DATA find_data;
      HANDLE hSearch = FindFirstFile( path, &find_data );

      if( hSearch != INVALID_HANDLE_VALUE ) {
         BOOL keep_going;

         do {
	    if( REGULAR_FILE_NAMEP( find_data.cFileName ) ) 
               res = MAKE_PAIR( string_to_bstring( find_data.cFileName ), res );
            keep_going= FindNextFile( hSearch, &find_data );
         } while( keep_going );

         FindClose( hSearch );
      }
   }
#endif
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_directory_to_path_list ...                                   */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_directory_to_path_list( char *name, int len, char sep ) {
   obj_t res = BNIL;
#ifndef _MSC_VER
   DIR *dir;
   struct dirent *dirent;

   if( (dir = opendir( name )) ) {
      while( (dirent = readdir( dir )) ) {
	 char *fname = dirent->d_name;

	 if( REGULAR_FILE_NAMEP( fname ) ) { 
	    obj_t bs = make_string_sans_fill( strlen( fname ) + len + 1 );
	    char *s = BSTRING_TO_STRING( bs );

	    strcpy( s, name );
	    s[ len ] = sep;
	    strcpy( s + len + 1, fname );

	    res = MAKE_PAIR( bs, res );
	 }
      }
      closedir( dir );
   }
#else
   char *const path = (char *)malloc( len + 2 + 1 );

   strcpy( path, name );
   strcat( path, "\\*" );

   {
      WIN32_FIND_DATA find_data;
      HANDLE hSearch = FindFirstFile( path, &find_data );

      if( hSearch != INVALID_HANDLE_VALUE ) {
         BOOL keep_going;

         do {
	    char *fname = find_data.cFileName; 
	    if( REGULAR_FILE_NAMEP( fname ) ) { 
	       obj_t bs = make_string_sans_fill( strlen( fname ) + len + 1 );
	       char *s = BSTRING_TO_STRING( bs );
	       strcpy( s, name );
	       s[ len ] = sep;
	       strcpy( s + len + 1, fname );
	       
	       res = MAKE_PAIR( bs, res );
	    }
            keep_going= FindNextFile( hSearch, &find_data );
         } while( keep_going );

         FindClose( hSearch );
      }
   }
#endif
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_file_to_string ...                                           */
/*---------------------------------------------------------------------*/
obj_t
bgl_file_to_string( char *path ) {
   int fd = open( path, O_RDONLY );

   if( !fd ) {
      C_SYSTEM_FAILURE( bglerror( errno ), 
			"file->string",
			strerror( errno ),
			string_to_bstring( path ) );
   } else {
      struct stat sin;
      
      if( fstat( fd, &sin ) ) {
	 close( fd );
	 C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			   "file->string",
			   strerror( errno ),
			   string_to_bstring( path ) );
      } else {
	 obj_t res = make_string_sans_fill( sin.st_size );
	 int n = read( fd, BSTRING_TO_STRING( res ), sin.st_size );

	 close( fd );
	 
	 if( sin.st_size != n ) {
	    C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			      "file->string",
			      strerror( errno ),
			      string_to_bstring( path ) );
	 } else {
	    close( fd );
	    return res;
	 }
      }
   }
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    copyfile ...                                                     */
/*    -------------------------------------------------------------    */
/*    This function is a replacement for sendfile. In particular       */
/*    it is used when the INPUT-PORT is a socket descriptor or when    */
/*    the host system does not support the sendfile system call.       */
/*---------------------------------------------------------------------*/
static long
copyfile( obj_t op, void *ip, long sz, long (*sysread)() ) {
   long rsz = 0, o;
   
   if( sz < 0 ) {
#ifdef __GNUC__
      char _buf[ default_io_bufsiz + 1 ];
      char *buf = _buf;
#else
      char *buf = alloca( default_io_bufsiz + 1 );
#endif
      long n;
      size_t m;

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile op=%p ip=%p sz=%d\n", op, ip, sz );
#endif
      
   loopr:
      while( (n = sysread( buf, 1, default_io_bufsiz, ip )) > 0 ) {
	 o = 0;
	 
	 bgl_write( op, &buf[ o ], n );
	 rsz += n;
      }

      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr;
      }

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout rsz=%d\n", rsz );
#endif

      bgl_output_flush( op, 0, 0 );
      return rsz;
   } else {
      long n = 0, m;
      long s = (default_io_bufsiz > sz) ? sz : default_io_bufsiz;
#ifdef __GNUC__
      char buf[ s + 1 ];
#else
      char *buf = alloca( s + 1 );
#endif

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout op=%p ip=%p sz=%d\n", op, ip, sz );
#endif
   loopr2:      
      while( (sz > 0) && (n = sysread( buf, 1, s, ip )) > 0 ) {
	 o = 0;
	 bgl_write( op, &buf[ o ], n );
	 rsz += n;
	 sz -= n;
	 
	 if( sz < s ) s = sz;
      }

      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr2;
      } 

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout rsz=%d\n", rsz );
#endif
      
      bgl_output_flush( op, 0, 0 );
      return rsz;
   }
}  

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gc_sendfile ...                                                  */
/*---------------------------------------------------------------------*/
#if( BGL_GC_HAVE_DO_BLOCKING )
static void
gc_sendfile( struct sendfile_info_t *si ) {
#if DEBUG_SENDCHARS
   fprintf( stderr, "gc_sendfile out=%p in=%p offset%d sz=%d\n",
	    si->out, si->in, si->offset, si->sz );
#endif
   si->res = _sendfile( si->out, si->in, si->offset, si->sz );
}
#endif

/*---------------------------------------------------------------------*/
/*    bgl_sendchars ...                                                */
/*    -------------------------------------------------------------    */
/*    uses sendfile to "copy" the input-port to the output-port        */
/*    flushes output-port!                                             */
/*---------------------------------------------------------------------*/
obj_t
bgl_sendchars( obj_t ip, obj_t op, long sz, long offset ) {
#define inp INPUT_PORT( ip )
#define outp OUTPUT_PORT( op )
   long dsz;
   long ws = 0;
   struct stat in;
   struct stat out;
   long n;

   if( (PORT( op ).kindof == KINDOF_CLOSED) ||
       (PORT( ip ).kindof == KINDOF_CLOSED) ||
       (PORT( ip ).kindof == KINDOF_GZIP) )
      return BFALSE;
      
   if( offset >= 0 ) {
      bgl_input_port_seek( ip, offset );
   } else {
      void *fd = PORT_STREAM( op );
      
      dsz = inp.bufpos - inp.matchstop - 1;

#if DEBUG
      fprintf( stderr, "%s:%d bgl_sendchars\n" );
#endif      
      bgl_output_flush( op, 0, 0 );

      if( dsz > 0 ) {
	 /* flush the input buffer when it contains characters */
	 size_t w;

	 ws = ((sz > 0) && (dsz > sz)) ? sz : dsz;
	 w = outp.syswrite( fd, &STRING_REF( inp.buf, inp.matchstop ), ws );
	 
#ifdef DEBUG_SENDCHARS   
	 fprintf( stderr, "bgl_sendchars.3: sz=%d offset=%d w=%d ws=%d\n", sz, offset, w, ws );
#endif	 
	 inp.matchstop += (long)w;
	 inp.forward = inp.matchstop;

	 if( w < ws ) {
	    C_SYSTEM_FAILURE( bglerror( errno ),
			      "send-chars",
			      strerror( errno ),
			      MAKE_PAIR( ip, op ) );
	    return BINT( w );
	 }

	 if( sz > 0 ) {
	    if( dsz >= sz ) {
#ifdef DEBUG_SENDCHARS   
	       fprintf( stderr, "bgl_sendchars.4: RETURN ws=%d\n", ws );
#endif	       
	       return BINT( ws );
	    }

	    sz -= ws;
	 }
      }
   }

#if( BGL_HAVE_SENDFILE )
   if( 
      /* Some operating systems (such as Linux 2.6.10) are demanding */
      /* on the input and output ports. These requirements are set   */
      /* in the configuration files and used to determine what has   */
      /* to be checked before invoking the actual sendfile sys call  */
#  if( BGL_SENDFILE_REQUIRE_INPUT_FILE )
       (inp.port.kindof != KINDOF_FILE)
       || fstat( fileno( (FILE *)PORT_STREAM( ip ) ), &in )
       || !S_ISREG( in.st_mode )
#  else
       0
#  endif
#  if( BGL_SENDFILE_REQUIRE_OUTPUT_SOCKET )
       || (outp.port.kindof != KINDOF_SOCKET)
       || _FSTAT( (int)PORT_STREAM( op ), &out )
       || ((out.st_mode & S_IFSOCK) != S_IFSOCK)
#  endif       
       || 0 ) {
#endif /* BGL_HAVE_SENDFILE */
      if( sz != 0 ) {
	 n = copyfile( op, ip, sz, INPUT_PORT( ip ).sysread );
      }
      if( n < 0 ) {
	 C_SYSTEM_FAILURE( bglerror( errno ), 
			   "send-chars",
			   strerror( errno ),
			   MAKE_PAIR( ip, op ) );
	 return BINT( 0 );
      }
#if( BGL_HAVE_SENDFILE )
   } else {
      if( sz == -1 ) sz = in.st_size;

      if( sz != 0 ) {
#if( BGL_GC_HAVE_BLOCKING )
	 bgl_gc_start_blocking();

	 n = _sendfile( (int)PORT_STREAM( outp ),
			fileno( (FILE *)PORT_STREAM( inp ) ),
			(offset > 0 ? (off_t *)(&offset) : 0),
			sz );
	 
	 bgl_gc_stop_blocking();
#else
#  if( BGL_GC_HAVE_DO_BLOCKING )
	 {
	    struct sendfile_info_t si;
	    si.out = (int)PORT_STREAM( op );
	    si.in = fileno( (FILE *)(PORT_STREAM( ip ) ) );
	    si.offset = (offset > 0 ? (off_t *)(&offset) : 0);
	    si.sz = sz;
	 
#ifdef DEBUG_SENDCHARS   
	    fprintf( stderr, "bgl_sendchars.5: sz=%d offset=%d\n", sz, offset );
#endif	    
	    bgl_gc_do_blocking( &gc_sendfile, &si );

	    n = si.res;
#ifdef DEBUG_SENDCHARS   
	    fprintf( stderr, "bgl_sendchars.6: n=%d\n", n );
#endif	    
	 }
      }
#  else
      -> error BGL_GC_HAVE_BLOCKING or BGL_GC_HAVE_DO_BLOCKING required
#  endif       
#endif /* BGL_HAVE_BLOCKING */
      if( n < 0 ) {
	 C_SYSTEM_FAILURE( bglerror( errno ),
			   "send-chars",
			   strerror( errno ),
			   MAKE_PAIR( ip, op ) );
      }
   }
   if( offset > 0 ) {
      if( INPUT_PORT_ON_FILEP( ip ) )
	 fseek( (FILE *)PORT_STREAM( ip ), offset + n + ws, SEEK_SET );
   }
#endif /* BGL_HAVE_SENDFILE */
   
   inp.filepos += n + ws;

   return BINT( n + ws );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_sendfile ...                                                 */
/*---------------------------------------------------------------------*/
obj_t
bgl_sendfile( obj_t name, obj_t op, long sz, long offset ) {
   struct stat sin;
   int in;
   _FD fd = (_FD)PORT_STREAM( op );
   int n;

   if( PORT( op ).kindof == KINDOF_CLOSED )
      return BFALSE;

      /* Some operating systems (such as Linux 2.6.10) are demanding */
      /* on the input and output ports. These requirements are set   */
      /* in the configuration files and used to determine what has   */
      /* to be checked before invoking the actual sendfile sys call  */
#  if( BGL_SENDFILE_REQUIRE_OUTPUT_SOCKET )
   if( outp.port.kindof != KINDOF_SOCKET)
      return BFALSE;
#  endif

   if( !(in = open( BSTRING_TO_STRING( name ), O_RDONLY, OMOD )) ) {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"send-file",
			strerror( errno ),
			name );
   }

   if( sz == -1 ) {
      if( fstat( in, &sin ) ) {
	 close( in );
	 C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			   "send-file",
			   strerror( errno ),
			   name );
      }
      sz = sin.st_size;
   }

#if DEBUG
      fprintf( stderr, "%s:%d bgl_sendfile\n" );
#endif      
   bgl_output_flush( op, 0, 0 );
   
#if( !BGL_HAVE_SENDFILE )
   n = copyfile( op, fdopen( in, "r" ), sz, (long (*)())&fread );
#else
   if( sz != 0 ) {
#  if( BGL_GC_HAVE_BLOCKING )
      bgl_gc_start_blocking();

      n = _sendfile( stream, in, (offset > 0 ? (off_t *)(&offset) : 0), sz );
	 
      bgl_gc_stop_blocking();
#  else
#    if( BGL_GC_HAVE_DO_BLOCKING )
      {
	 struct sendfile_info_t si;
	 si.out = fd;
	 si.in = in;
	 si.offset = (offset > 0 ? (off_t *)(&offset) : 0);
	 si.sz = sz;

	 bgl_gc_do_blocking( &gc_sendfile, &si );

	 n = si.res;
      }
#    else
      -> error BGL_GC_HAVE_BLOCKING or BGL_GC_HAVE_DO_BLOCKING required
#    endif       
#  endif /* BGL_HAVE_BLOCKING */

   }
#endif
   
   close( in );
      
   if( n < 0 )
      C_SYSTEM_FAILURE( bglerror( errno ),
			"send-file",
			strerror( errno ),
			MAKE_PAIR( name, op ) );
   return BINT( n );
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    strwrite ...                                                     */
/*---------------------------------------------------------------------*/
static size_t
strwrite( obj_t port, void *str, size_t count ) {
   obj_t buf = OUTPUT_PORT( port ).buf;
   long cnt = OUTPUT_PORT( port ).cnt;
   long used = STRING_LENGTH( buf ) - cnt;
   long nlen = (STRING_LENGTH( buf ) + count) * 2;
   obj_t nbuf = make_string_sans_fill( nlen );

#if DEBUG   
   fprintf( stderr, "%s:%d>>> strwrite count=%d\n", __FILE__, __LINE__, count );
#endif

   memcpy( &STRING_REF( nbuf, 0 ), &STRING_REF( buf, 0 ), used );
   memcpy( &STRING_REF( nbuf, used ), str, count );

   OUTPUT_PORT( port ).ptr = &STRING_REF( nbuf, used + count );
   OUTPUT_PORT( port ).cnt = nlen - (used + count);
   OUTPUT_PORT( port ).buf = nbuf;

#if DEBUG   
   fprintf( stderr, "%s:%d<<< strwrite nlen=%d cnt=%d count=%d\n",
	    __FILE__, __LINE__,
	    nlen, OUTPUT_PORT( port ).cnt, count );
#endif
   
   return count;
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    procwrite ...                                                    */
/*---------------------------------------------------------------------*/
static size_t
procwrite( obj_t port, void *str, size_t sz ) {
   obj_t proc = VECTOR_REF( PORT( port ).userdata, 0 );
   obj_t buf = VECTOR_REF( PORT( port ).userdata, 1 );
   int len = STRING_LENGTH( buf );

   if( sz > len ) {
      buf = make_string_sans_fill( sz + 1 );
      len = sz + 1;
      VECTOR_SET( PORT( port ).userdata, 1, buf );
   }
      
   memcpy( &STRING_REF( buf, 0 ), str, sz );
   STRING_SET( buf, sz, '\0' );
   STRING_LENGTH( buf ) = sz;
   
   PROCEDURE_ENTRY( proc )( proc, buf, BEOA );
   
   STRING_LENGTH( buf ) = len;
   return sz;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    procflush ...                                                    */
/*---------------------------------------------------------------------*/
static obj_t
procflush( obj_t port ) {
   obj_t flush = VECTOR_REF( PORT( port ).userdata, 3 );

   return PROCEDURE_ENTRY( flush )( flush, BEOA );
}

/*---------------------------------------------------------------------*/
/*    BGL_RUNTIME_DEF obj_t                                            */
/*    bgl_password ...                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_password( char *prompt ) {
   int max_len = 80;
   char *s = alloca( max_len + 1 );
   int i = 0;
   int c;
   FILE *tty = fopen( "/dev/tty", "w" );
   FILE *out = tty ? tty : stderr;
   
#if HAVE_TERMIO || HAVE_TERMIOS
   struct termios t;
   tcflag_t lflag;

   fputs( prompt, out );
   fflush( out );

   /* set the terminal in no echo mode */
   tcgetattr( 0, &t );
   lflag = t.c_lflag;
   t.c_lflag &= ~(ICANON|ECHO);
   t.c_cc[ VMIN ] = 1;
   t.c_cc[ VTIME ] = 0;
   tcsetattr( 0, TCSANOW, &t );
#endif

   while( (c = getchar()) != '\n' ) {
      if( i == max_len ) {
	 char *ns = alloca( max_len * 2 );
	 memcpy( ns, s, max_len );
	 max_len *= 2;
	 s = ns;
      }
      
      s[ i++ ] = c;
      putc( '*', out );
      fflush( out );
   }

   s[ i ] = 0;
   
#if HAVE_TERMIO || HAVE_TERMIOS
   /* restore the terminal */
   t.c_lflag = lflag;
   tcsetattr( 0, TCSANOW, &t );
#endif

   putc( '\n', out );
   fflush( out );
   
   if( tty ) fclose( tty );
   
   return string_to_bstring_len( s, i );
}

