/*=====================================================================*/
/*    serrano/prgm/project/bigloo/runtime/Clib/csocket.c               */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Jun 29 18:18:45 1998                          */
/*    Last change :  Fri Sep 12 15:34:23 2008 (serrano)                */
/*    -------------------------------------------------------------    */
/*    Scheme sockets                                                   */
/*    -------------------------------------------------------------    */
/*    This file is based on a contribution of                          */
/*    David Tolpin (dvd@pizza.msk.su)                                  */
/*                                                                     */
/*    Bugs correction (conversion between host and network byte order) */
/*    by Marc Furrer (Marc.Furrer@di.epfl.ch)                          */
/*                                                                     */
/*    Reworked  by Erick Gallesio for 2.2 release.                     */
/*    Some additions and simplifications (I hope).                     */
/*                                                                     */
/*    DNS caching added by Manuel Serrano, 22 oct 2006.                */
/*=====================================================================*/
#if defined( _MSC_VER) || defined( _MINGW_VER )
#  define _BGL_WIN32_VER
#endif

#include <bigloo_config.h>
#include <sys/types.h>
#include <time.h>
#ifndef _BGL_WIN32_VER
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include <netdb.h>
#   if( BGL_HAVE_SELECT )
#     include <sys/time.h>
#     include <sys/types.h>
#     include <unistd.h>
#   endif
#else
#   if defined( _MINGW_VER )
#      include "windows.h"
#   endif
#   include <winsock2.h>
#   include <mswsock.h>
#   include <io.h>
#endif
#include <fcntl.h>
#include <memory.h>
#include <errno.h>
#include <bigloo.h>

#ifndef _BGL_WIN32_VER
#   define BAD_SOCKET(s) ((s) < 0)
#else
#   define BAD_SOCKET(s) ((s) == INVALID_SOCKET)
#endif

#if !BGL_HAVE_SOCKLEN
typedef int socklen_t;
#endif

#if( BGL_HAVE_SOCKET_TCP_NODELAY || BGL_HAVE_SOCKET_TCP_CORK )
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
static int sock_opt = 1;
#endif

/*---------------------------------------------------------------------*/
/*    Imports ...                                                      */
/*---------------------------------------------------------------------*/
extern obj_t bgl_close_input_port( obj_t );
extern obj_t bgl_file_to_buffered_input_port( obj_t, FILE *, obj_t );
extern long bgl_pipe_fread( char *, long, long, obj_t );
extern obj_t make_vector();
extern unsigned char get_hash_number( char * );
extern unsigned char bgl_get_hash_number_len( char *, int, int );
extern bool_t bigloo_strcmp( obj_t o1, obj_t o2 );
extern bool_t bgl_dns_enable_cache();
extern long bgl_dns_cache_validity_timeout();

#ifndef _BGL_WIN32_VER
extern int dup( int );
extern int close( int );
#endif

/*---------------------------------------------------------------------*/
/*    bglhostent ...                                                   */
/*---------------------------------------------------------------------*/
struct bglhostent {
   header_t header;
   struct hostent hp;
   obj_t hostaddr;
   long exptime;
};

/*---------------------------------------------------------------------*/
/*    socket mutex                                                     */
/*---------------------------------------------------------------------*/
static obj_t socket_mutex = BUNSPEC;
DEFINE_STRING( socket_mutex_name, _2, "socket-mutex", 12 );
static obj_t gethostby_mutex = BUNSPEC;
DEFINE_STRING( gethostby_name, _3, "socket-gethostby", 22 );
static obj_t socket_port_mutex = BUNSPEC;
DEFINE_STRING( socket_port_mutex_name, _4, "socket-port-mutex", 17 );

/*---------------------------------------------------------------------*/
/*    Global C variables                                               */
/*---------------------------------------------------------------------*/
static obj_t hosttable = BUNSPEC;
static obj_t addrtable = BUNSPEC;

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_socket ...                                              */
/*    -------------------------------------------------------------    */
/*    Initialize the socket machinery (mostly host names caching).     */
/*---------------------------------------------------------------------*/
void
bgl_init_socket() {
   if( !VECTORP( hosttable ) ) {
      socket_mutex = bgl_make_mutex( socket_mutex_name );
      gethostby_mutex = bgl_make_mutex( gethostby_name );
//      socket_port_mutex = bgl_make_mutex( socket_port_mutex_name );
      hosttable = make_vector( 256, 0 );
      addrtable = make_vector( 256, 0 );
   }
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    socket_error ...                                                 */
/*---------------------------------------------------------------------*/
static void
socket_error( char *who, char *message, obj_t object ) {
   C_SYSTEM_FAILURE( BGL_IO_ERROR, who, message, object );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    socket_timeout_error ...                                         */
/*---------------------------------------------------------------------*/
static void
socket_timeout_error( obj_t hostname, int port ) {
   char buffer[ 512 ];
   
   sprintf( buffer, "%s:%d", BSTRING_TO_STRING( hostname ), port );

   C_SYSTEM_FAILURE( BGL_IO_TIMEOUT_ERROR,
		     "make-client-socket",
		     "Connection time out",
		     string_to_bstring( buffer ) );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    client_socket_error ...                                          */
/*---------------------------------------------------------------------*/
static void
client_socket_error( obj_t hostname, int port, char *msg ) {
   char buffer1[ 512 ];
   char buffer2[ 512 ];

   if( msg ) sprintf( buffer1, "%s, ", msg );
   sprintf( buffer1, "%s (%d)", strerror( errno ), errno );
   
   sprintf( buffer2, "%s:%d", BSTRING_TO_STRING( hostname ), port );

   socket_error( "make-client-socket", buffer1, string_to_bstring( buffer2 ) );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    system_error ...                                                 */
/*---------------------------------------------------------------------*/
static void
system_error( char *who, obj_t val ) {
   char buffer[ 512 ];

   sprintf( buffer, "%s (%d)", strerror( errno ), errno );

   socket_error( who, buffer, val );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    make_string ...                                                  */
/*---------------------------------------------------------------------*/
static char *
make_string( char *src ) {
   char *res = (char *)GC_MALLOC_ATOMIC( strlen( src ) + 1 );

   strcpy( res, src );
   return res;
}

/*---------------------------------------------------------------------*/
/*    static char **                                                   */
/*    make_string_array ...                                            */
/*---------------------------------------------------------------------*/
static char **
make_string_array( char **src ) {
   int len = 0;
   char **run, **res;

   for( run = src; *run; run++ );
   len = (run - src);

   res = (char **)GC_MALLOC( sizeof( char * ) * len + 1 );
   for( run = src; *run; run++ ) {
      *res++ = make_string( *run );
   }

   *res = 0;

   return res - len;
}

/*---------------------------------------------------------------------*/
/*    static char **                                                   */
/*    make_inet_array ...                                              */
/*---------------------------------------------------------------------*/
static char **
make_inet_array( char **src, int size ) {
   int len = 0;
   char **run, **res;

   for( run = src; *run; run++ );
   len = (run - src);

   res = (char **)GC_MALLOC( sizeof( char * ) * len + 1 );
   for( run = src; *run; run++ ) {
      unsigned char *d = (unsigned char *)GC_MALLOC_ATOMIC( size );
      unsigned char *s = *run;
      memcpy( d, s, size );
      *res++ = d;
   }

   *res = 0;

   return res - len;
}

/*---------------------------------------------------------------------*/
/*    static struct bglhostent *                                       */
/*    make_bglhostent ...                                              */
/*---------------------------------------------------------------------*/
static struct bglhostent *
make_bglhostent( obj_t hostaddr, struct hostent *hp ) {
   struct bglhostent *bhp = (struct bglhostent *)GC_MALLOC( sizeof( struct bglhostent ) );

   bhp->header = MAKE_HEADER( OPAQUE_TYPE, 0 );
   bhp->hostaddr = hostaddr;
   bhp->exptime = time( 0L ) + bgl_dns_cache_validity_timeout();
   memcpy( &(bhp->hp), hp, sizeof( struct hostent ) );
   bhp->hp.h_name = make_string( hp->h_name );
   bhp->hp.h_aliases = make_string_array( hp->h_aliases );
   bhp->hp.h_addr_list = make_inet_array( hp->h_addr_list, hp->h_length );

   return bhp;
}

/*---------------------------------------------------------------------*/
/*    static struct bglhostent *                                       */
/*    make_bglhostentbyname ...                                        */
/*---------------------------------------------------------------------*/
static struct bglhostent *
make_bglhostentbyname( obj_t hostname ) {
   struct hostent *hp;
   
   bgl_mutex_lock( gethostby_mutex );
   hp = gethostbyname( BSTRING_TO_STRING( hostname ) );
   if( !hp ) {
      bgl_mutex_unlock( gethostby_mutex );
      return 0L;
   } else {
      struct bglhostent *res = make_bglhostent( hostname, hp );
      bgl_mutex_unlock( gethostby_mutex );
      return res;
   }
}

static struct bglhostent *
make_bglhostentbyname_dbg( obj_t hostname ) {
   fprintf( stderr, "%s:%d make_bglgethostbyname: %s\n", __FILE__, __LINE__,
	    BSTRING_TO_STRING( hostname ) );
   return make_bglhostentbyname( hostname );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    invalidate_hostbyname ...                                        */
/*---------------------------------------------------------------------*/
static void
invalidate_hostbyname( obj_t hostname ) {
#if BGL_DNS_CACHE
   if( bgl_dns_enable_cache() ) {
      int key = get_hash_number( BSTRING_TO_STRING( hostname ) );
      struct bglhostent *bhp;
      
      bgl_mutex_lock( socket_mutex );
      
      bhp = (struct bglhostent *)VECTOR_REF( hosttable, key );
/*       fprintf( stderr, "INVALIDATE_HOSTBYNAME: %d\n", key );        */
      if( bhp && bigloo_strcmp( bhp->hostaddr, hostname ) ) {
	 VECTOR_SET( hosttable, key, 0L );
      }
      
      bgl_mutex_unlock( socket_mutex );
      
      return;
   }
#endif
}

/*---------------------------------------------------------------------*/
/*    struct hostent *                                                 */
/*    bglhostbyname ...                                                */
/*    -------------------------------------------------------------    */
/*    It function is comparable to gethostbyname except that           */
/*    it accepts a Bigloo string instead of a C string.                */
/*                                                                     */
/*    This function is thread-safe. It can be called simultaneously    */
/*    by several threads. It returns a fresh data-structure so         */
/*    client don't have to deploy a lock machinery for using it.       */
/*---------------------------------------------------------------------*/
static struct hostent *
bglhostbyname( obj_t hostname ) {
   struct bglhostent *bhp;
#if BGL_DNS_CACHE
   if( bgl_dns_enable_cache() ) {
      int key = get_hash_number( BSTRING_TO_STRING( hostname ) );

      /* acquire the global socket lock */
      bgl_mutex_lock( socket_mutex );

      bhp = (struct bglhostent *)VECTOR_REF( hosttable, key );

      /* is it currently in the table */
      if( bhp
	  && bigloo_strcmp( bhp->hostaddr, hostname )
	  && ((time( 0 ) - bhp->exptime) <= 0) ) {
	 bgl_mutex_unlock( socket_mutex );

	 return &(bhp->hp);
      } else {
	 bgl_mutex_unlock( socket_mutex );
	 
	 if( bhp = make_bglhostentbyname( hostname ) ) {
	    
	    bgl_mutex_lock( socket_mutex );
	    VECTOR_SET( hosttable, key, (obj_t)bhp );
	    bgl_mutex_unlock( socket_mutex );
	    
	    return &(bhp->hp);
	 } else {
	    return 0L;
	 }
      }
   } else 
#endif
   {
      bhp = make_bglhostentbyname( hostname );
      
      if( bhp )
	 return &(bhp->hp);
      else
	 return 0L;
   }
}

/*---------------------------------------------------------------------*/
/*    static struct bglhostent *                                       */
/*    make_bglhostentbyaddr ...                                        */
/*---------------------------------------------------------------------*/
static struct bglhostent *
make_bglhostentbyaddr( obj_t hostaddr, struct sockaddr_in *sin ) {
   struct hostent *hp;

   bgl_mutex_lock( gethostby_mutex );
   hp = gethostbyaddr( (char *)&(sin->sin_addr),
		       sizeof( sin->sin_addr ),
		       AF_INET );

   if( !hp ) {
      bgl_mutex_unlock( gethostby_mutex );
      return 0L;
   } else {
      struct bglhostent *res = make_bglhostent( hostaddr, hp );
      bgl_mutex_unlock( gethostby_mutex );
      return res;
   }
}

static struct bglhostent *
make_bglhostentbyaddr_dbg( obj_t hostaddr, struct sockaddr_in *sin ) {
   fprintf( stderr, "%s:%d make_bglgethostbyaddr: %d.%d.%d.%d\n", __FILE__, __LINE__,
	    ((unsigned char *)BSTRING_TO_STRING( hostaddr ))[ 0 ],
	    ((unsigned char *)BSTRING_TO_STRING( hostaddr ))[ 1 ],
	    ((unsigned char *)BSTRING_TO_STRING( hostaddr ))[ 2 ],
	    ((unsigned char *)BSTRING_TO_STRING( hostaddr ))[ 3 ] );
   return make_bglhostentbyaddr( hostaddr, sin );
}

/*---------------------------------------------------------------------*/
/*    struct hostent *                                                 */
/*    bglhostbyaddr ...                                                */
/*    -------------------------------------------------------------    */
/*    See bglhostbynadd.                                               */
/*---------------------------------------------------------------------*/
static struct hostent *
bglhostbyaddr( struct sockaddr_in *sin ) {
   struct bglhostent *bhp;

#if BGL_DNS_CACHE
   if( bgl_dns_enable_cache() ) {
      int key = bgl_get_hash_number_len( (char *)&(sin->sin_addr),
					 0,
					 sizeof( sin->sin_addr ) );

      /* acquire the global socket lock */
      bgl_mutex_lock( socket_mutex );

      bhp = (struct bglhostent *)VECTOR_REF( addrtable, key );

      /* is it currently in the table */
      if( bhp
	  && !strncmp( BSTRING_TO_STRING( bhp->hostaddr ),
		      (char *)&(sin->sin_addr),
		      sizeof( sin->sin_addr ) )
	  && ((time( 0 ) - bhp->exptime) <= 0) ) {
	 bgl_mutex_unlock( socket_mutex );

	 return &(bhp->hp);
      } else {
	 obj_t hostaddr = string_to_bstring_len( (char *)&(sin->sin_addr),
						 sizeof( sin->sin_addr ) );
	 bgl_mutex_unlock( socket_mutex );

	 if( bhp = make_bglhostentbyaddr( hostaddr, sin ) ) {
	    
	    bgl_mutex_lock( socket_mutex );
	    VECTOR_SET( addrtable, key, (obj_t)bhp );
	    bgl_mutex_unlock( socket_mutex );
	    
	    return &(bhp->hp);
	 } else {
	    return 0L;
	 }
      }
   } else 
#endif
   {
      obj_t hostaddr = string_to_bstring_len( (char *)&(sin->sin_addr),
					    sizeof( sin->sin_addr ) );
      bhp = make_bglhostentbyaddr( hostaddr, sin );
      
      if( bhp )
	 return &(bhp->hp);
      else
	 return 0L;
   }
}

/*---------------------------------------------------------------------*/
/*    static struct hostent *                                          */
/*    bgl_gethostent ...                                               */
/*---------------------------------------------------------------------*/
static struct hostent *
bgl_gethostent( obj_t hostname ) {
   struct hostent *hp;

   if( (hp = bglhostbyname( hostname )) == NULL ) {
      char *msg;
      
      switch( h_errno ) {
         case HOST_NOT_FOUND:
	    msg = "Unknown host";
	    break;
	 case NO_ADDRESS:
	    msg = "No address or no data";
	    break;
	 case NO_RECOVERY:
	    msg = "Internal DNS error";
	    break;
	 case TRY_AGAIN:
	    msg = "temporary error";
	    break;
	 default:
	    msg = "Unknown error";
      }
	    
      C_SYSTEM_FAILURE( BGL_IO_UNKNOWN_HOST_ERROR, "host", msg, hostname );
   }

   return hp;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_host ...                                                     */
/*---------------------------------------------------------------------*/
obj_t
bgl_host( obj_t hostname ) {
   struct hostent *hp = bgl_gethostent( hostname );
   
   return string_to_bstring( inet_ntoa( *(struct in_addr *)(hp->h_addr) ) );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_hostinfo ...                                                 */
/*---------------------------------------------------------------------*/
obj_t
bgl_hostinfo( obj_t hostname ) {
   struct hostent *hp = bgl_gethostent( hostname );
   obj_t res = BNIL;
   obj_t addr = BNIL;
   obj_t alias = BNIL;
   obj_t name = BNIL;
   char **runner;
   obj_t s;

   for( runner = hp->h_addr_list; *runner; runner++ ) {
      s = string_to_bstring( inet_ntoa( *(struct in_addr *)(*runner) ) );
      addr = MAKE_PAIR( s, addr );
   }

   for( runner = hp->h_aliases; *runner; runner++ ) {
      alias = MAKE_PAIR( string_to_bstring( *runner ), alias );
   }

   if( PAIRP( alias ) ) {
      alias = MAKE_PAIR( string_to_symbol( "aliases" ), alias );
      res = MAKE_PAIR( alias, res );
   }
   if( PAIRP( addr ) ) {
      addr = MAKE_PAIR( string_to_symbol( "addresses" ), addr );
      res = MAKE_PAIR( addr, res );
   }

   s = string_to_bstring( hp->h_name );
   name = MAKE_PAIR( s, BNIL );
   name = MAKE_PAIR( string_to_symbol( "name"), name );
		     
   res = MAKE_PAIR( name, res );

   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_gethostname ...                                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_gethostname() {
#define MAXHOSTNAME 1024   
   struct hostent *hp;
   static char h[ MAXHOSTNAME + 1 ];
   obj_t res;
   
   gethostname( h, MAXHOSTNAME );
   hp = bglhostbyname( string_to_bstring( h ) );

   res = string_to_bstring( hp ? hp->h_name : "localhost" );

   return res;
#undef MAXHOSTNAME
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    socket_startup ...                                               */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
socket_startup() {
#ifdef _BGL_WIN32_VER
   WSADATA data;
   int result = 0;
   WORD version_requested = MAKEWORD( 2, 2 );
   DWORD val = SO_SYNCHRONOUS_NONALERT;

   result = WSAStartup( version_requested, &data );
   if( 0 != result ) {
	socket_error( "socket_init",
		      "Failed to Initialize socket library",
		      BNIL );
   }

   result = setsockopt( INVALID_SOCKET,
			SOL_SOCKET, SO_OPENTYPE,
			(const char *)&val,
			sizeof( val ) );
   if( 0 != result ) {
      socket_error( "make_server_socket",
		    "cannot set socket options",
		    BUNSPEC);
   }
#endif
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    socket_cleanup ...                                               */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
socket_cleanup() {
#ifdef _BGL_WIN32_VER
   WSACleanup();
#endif
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    bgl_flushwrite ...                                               */
/*---------------------------------------------------------------------*/
static size_t
bgl_flushwrite( void *ptr, size_t size, size_t nmemb, obj_t op ) {
   size_t r = fwrite( ptr, size, nmemb, PORT_STREAM( op ) );
   fflush( PORT_STREAM( op ) );
   return r;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_flushputc ...                                                */
/*---------------------------------------------------------------------*/
static int
bgl_flushputc( int c, obj_t op ) {
   int r = putc( c, PORT_STREAM( op ) );
   fflush( PORT_STREAM( op ) );
   return r;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_sclose_rd ...                                                */
/*---------------------------------------------------------------------*/
static int
bgl_sclose_rd( FILE *stream ) {
#if( !defined( SHUT_RD ) )
#  define SHUT_RD 1
#endif   
   shutdown( fileno( stream ), SHUT_RD );
   return fclose( stream );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    set_socket_io_ports ...                                          */
/*---------------------------------------------------------------------*/
static void
set_socket_io_ports( int s, obj_t sock, char *who, obj_t inb, obj_t outb ) {
   int t, port;
   obj_t host;
   FILE *fs;

   /* if on windows obtain a C run-time compatible file descriptor */
#ifdef _BGL_WIN32_VER
   s = _open_osfhandle( s, _O_RDWR );
#endif

   /* duplicate handles so that we are able to access one */
   /* socket channel via two scheme ports.                */
   t = dup( s );

   if( (t == -1) || (s == -1) ) {
      char *buffer = alloca( 1024 );
      
      sprintf( buffer, "%s: cannot duplicate io port, %s",
	       who, strerror( errno ) );
      socket_error( "set_socket_io_ports", buffer, BUNSPEC );
   }

   if( !(fs = fdopen( s, "r" )) ) { 
      char *buffer = alloca( 1024 );
      
      sprintf( buffer, "%s: cannot create socket io ports, %s (s=%d->%p)",
	       who, strerror( errno ), s, fs );
      socket_error( "set_socket_io_ports", buffer, sock );
   }

   port = SOCKET( sock ).portnum;
   host = SOCKET( sock ).hostip;

   /* Create input port */
   SOCKET( sock ).input = bgl_file_to_buffered_input_port( host, fs, inb );
   SOCKET( sock ).input->port_t.kindof = KINDOF_SOCKET;
   SOCKET( sock ).input->input_port_t.sysread = bgl_pipe_fread;
   SOCKET( sock ).input->port_t.sysclose = &bgl_sclose_rd;

   /* Create output port */
   SOCKET( sock ).output = bgl_make_output_port( host, (void *)t,
						 KINDOF_SOCKET,
						 outb,
						 (size_t (*)())write,
						 lseek,
						 close );

   if( STRING_LENGTH( inb ) <= 1 )
      OUTPUT_PORT( SOCKET( sock ).output ).syswrite = bgl_flushwrite;

   if( STRING_LENGTH( outb ) <= 1 )
      OUTPUT_PORT( SOCKET( sock ).output ).bufmode = BGL_IONB;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    set_socket_blocking ...                                          */
/*---------------------------------------------------------------------*/
#if( BGL_HAVE_FCNTL )
static void
set_socket_blocking( int fd, int bool ) {
   int val;

   if( (val = fcntl( fd, F_GETFL, 0)) < 0 ) {
      socket_error( "make_client_socket", "cannot get socket control", BUNSPEC);
   }

   if( bool ) {
      val |= O_NONBLOCK;
   } else {
      val &= ~O_NONBLOCK;
   }

   if( fcntl( fd, F_SETFL, val ) < 0) {
      socket_error( "make_client_socket", "cannot set socket control", BUNSPEC);
   }
}
#endif

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_make_client_socket ...                                       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_make_client_socket( obj_t hostname, int port, int timeo, obj_t inb, obj_t outb ) {
   struct hostent *hp;
   struct sockaddr_in server;
   int s, err;
   obj_t a_socket;
   obj_t hname;

   /* Locate the host IP address */
   if( (hp = bglhostbyname( hostname )) == NULL ) {
      C_SYSTEM_FAILURE( BGL_IO_UNKNOWN_HOST_ERROR,
			"make-client-socket",
			"unknown or misspelled host name",
			hostname );
   }

   /* Get a socket */
   if( BAD_SOCKET( s = (int)socket( AF_INET, SOCK_STREAM, 0 ) ) ) {
      client_socket_error( hostname, port, "cannot create socket" );
   }

   /* Setup a connect address */
   memset( &server, 0, sizeof( server ) );
   memcpy( (char *)&server.sin_addr, hp->h_addr, hp->h_length );
   server.sin_family = AF_INET;
   server.sin_port = htons( port );

   hname = string_to_bstring( hp->h_name );
   
#if( BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   if( timeo > 0 ) set_socket_blocking( s, 1 );
#endif

   
   /* Try to connect */
   while( (err = connect( s,
			  (struct sockaddr *)&server,
			  sizeof(struct sockaddr_in)) ) != 0
          && errno == EINTR);
   
   if( err < 0 ) {
#if( BGL_HAVE_SELECT && defined( EINPROGRESS ) ) 
      if( errno == EINPROGRESS ) {
	 fd_set writefds;
	 struct timeval timeout;

	 FD_ZERO( &writefds );
	 FD_SET( s, &writefds );
	 timeout.tv_sec = timeo / 1000000;;
	 timeout.tv_usec = timeo % 1000000;

         /* retry if interrupted */
         while((err = select( s + 1, NULL, &writefds, NULL, &timeout )) < 0
               && errno == EINTR);

	 if( err == 0 ) {
	    /* we have experienced a failure so we */
	    /* invalidate the host name entry */
	    invalidate_hostbyname( hostname );
      
	    close( s );
	    socket_timeout_error( hostname, port );
	 } else {
	    if( err < 0 ) {
	       /* we have experienced a failure so we */
	       /* invalidate the host name entry */
	       invalidate_hostbyname( hostname );
	       
	       close( s );
	       client_socket_error( hostname, port, "Connection failed" );
	    } else {
	       int len = sizeof( int );
	       int r = getsockopt( s, SOL_SOCKET, SO_ERROR, (void *)&err, (socklen_t *)&len );

	       if( (r < 0) || (err != 0) ) {
		  /* we have experienced a failure so we */
		  /* invalidate the host name entry */
		  invalidate_hostbyname( hostname );
      
		  close( s );
		  client_socket_error( hostname, port, 0 );
	       }
	    }
	 }
	 set_socket_blocking( s, 0 );
      } else {
	 /* we have experienced a failure so we */
	 /* invalidate the host name entry */
	 invalidate_hostbyname( hostname );
      
	 close( s );
	 client_socket_error( hostname, port, "Connection failed" );
      }
#else
      /* we have experienced a failure so we */
      /* invalidate the host name entry */
      invalidate_hostbyname( hostname );
      
      close( s );
      client_socket_error( hostname, port, "Connection failed" );
#endif
   }

   /* Tune the server socket */
#if( BGL_HAVE_SOCKET_TCP_CORK )
   setsockopt( s, IPPROTO_TCP, TCP_CORK, &sock_opt, sizeof( sock_opt ) );
#else   
#  if( BGL_HAVE_SOCKET_TCP_NODELAY )
   setsockopt( s, IPPROTO_TCP, TCP_NODELAY, &sock_opt, sizeof( sock_opt ) );
#  endif
#endif
   
   /* Create a new Scheme socket object */
   a_socket = GC_MALLOC( SOCKET_SIZE );
   a_socket->socket_t.header = MAKE_HEADER( SOCKET_TYPE, 0 );
   a_socket->socket_t.portnum = ntohs( server.sin_port );
   a_socket->socket_t.hostname = hname;
   a_socket->socket_t.hostip = string_to_bstring( inet_ntoa( server.sin_addr ) );
   a_socket->socket_t.fd = s;
   a_socket->socket_t.input = BFALSE;
   a_socket->socket_t.output = BFALSE;
   a_socket->socket_t.stype = SOCKET_CLIENT;
   a_socket->socket_t.userdata = BUNSPEC;

   set_socket_io_ports( s, BREF( a_socket ), "make-client-socket", inb, outb );
   
   return BREF( a_socket );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_make_server_socket ...                                       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_make_server_socket( obj_t hostname, int port, int backlog ) {
   char msg[] = "make-server-socket";
   struct sockaddr_in sin;
   struct hostent *hp;
   int s, portnum = port, len;
   obj_t a_socket;

   /* Determine port to use */
   if( portnum < 0 )
      socket_error( "make-server-socket", "bad port number", BINT( port ) );

   /* Locate the host IP address */
   if( (hostname != BFALSE) && !(hp = bglhostbyname( hostname )) ) {
     socket_error( "make-server-socket", "unknown or misspelled host name", 
                   hostname );
   }

   /* Create a socket */
   if( BAD_SOCKET(s = (int)socket( AF_INET, SOCK_STREAM, 0 )) ){
      socket_error( "make-server-socket", "Cannot create socket", BUNSPEC );
   }

/*    {                                                                */
/*       int fd = s;                                                   */
/*       int flags;                                                    */
/*       {* set the socket in non-blocking mode *}                     */
/*       flags = fcntl( fd, F_GETFL );                                 */
/*       if( flags == -1 ) {                                           */
/* 	 return 0;                                                     */
/*       }                                                             */
/*                                                                     */
/*       flags |= O_NONBLOCK;                                          */
/*       flags = fcntl( fd, F_SETFL, flags );                          */
/*                                                                     */
/*       if( flags == -1 ) {                                           */
/* 	 return 0;                                                     */
/*       }                                                             */
/*    }                                                                */
   
   /* Bind the socket to a name */
   if( hostname != BFALSE ) {
      memset( &sin, 0, sizeof( sin ) );
      memcpy( (char*)&sin.sin_addr, hp->h_addr, hp->h_length );
   } else {
     sin.sin_addr.s_addr = INADDR_ANY;
   }

   sin.sin_family = AF_INET;
   sin.sin_port = htons( portnum );

   /* OPTIONAL */
   {
      int optval = 1;

      /* set the reuse flag */
      if( setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		     (char *) &optval, sizeof(optval)) < 0 ) {
	 system_error( msg, BINT( port ) );
      }
   }

   if( bind( s, (struct sockaddr *)&sin, sizeof( sin ) ) < 0 ) {
      close( s );
      system_error( msg, BINT( port ) );
   }

   /* Query the socket name, permits to get the true socket number */
   /* if 0 was given                                               */
   len = sizeof( sin );
   if( getsockname( s, (struct sockaddr *) &sin, (socklen_t *) & len ) < 0 ) {
      close( s );
      system_error( msg, BINT( port ) );
   }

   /* Indicate that we are ready to listen */
   if( listen( s, backlog ) < 0 ) {
      close( s );
      system_error( msg, BINT( port ) );
   }

   /* Now we can create the socket object */
   a_socket = GC_MALLOC( SOCKET_SIZE );
   a_socket->socket_t.header = MAKE_HEADER( SOCKET_TYPE, 0 );
   a_socket->socket_t.portnum = ntohs( sin.sin_port );
   a_socket->socket_t.hostname = BUNSPEC;
   a_socket->socket_t.hostip = BFALSE;
   a_socket->socket_t.fd = s;
   a_socket->socket_t.input = BFALSE;
   a_socket->socket_t.output = BFALSE;
   a_socket->socket_t.stype = SOCKET_SERVER;
   a_socket->socket_t.accept = 0L;
   a_socket->socket_t.userdata = BUNSPEC;

   return BREF( a_socket );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    socket_local_addr ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
socket_local_addr( obj_t sock ) {
   struct sockaddr_in sin;
   int len = sizeof( sin );

   if( SOCKET( sock ).stype == SOCKET_SERVER ) {
      return string_to_bstring( "0.0.0.0" );
   }

   if( getsockname( SOCKET( sock ).fd,
		    (struct sockaddr *)&sin,
		    (socklen_t *) & len) )
      socket_error( "socket-local-address", strerror( errno ), sock );

   return string_to_bstring( (char *)inet_ntoa( sin.sin_addr ) );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    socket_shutdown ...                                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
socket_shutdown( obj_t sock, int close_socket ) {
   int fd = SOCKET( sock ).fd;
   obj_t chook = SOCKET_CHOOK( sock );

   if( fd > 0 ) {
      /* MS: 19 Aug 2008, we don't have to close fd */
      /* since it will be closed automatically with */
      /* SOCKET( sock ).input                       */
      SOCKET( sock ).fd = -1;
      
      if( close_socket ) {
#if( !defined( SHUT_RDWR ) )
#  define SHUT_RDWR 2
#endif   
	 if( shutdown( fd, SHUT_RDWR ) ) {
	    char *buffer = alloca( 1024 );
	    
	    sprintf( buffer, "cannot shutdown socket, %s", strerror( errno ) );
	    socket_error( "socket-shutdown", buffer, sock );
	 }
      }

      if( PROCEDUREP( chook ) ) {
	 if( PROCEDURE_ARITY(chook) == 1 ) {
	    PROCEDURE_ENTRY( chook )( chook, sock, BEOA );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "socket-shutdown",
			      "illegal close hook arity",
			      chook );
	 }
      }

      /* Warning: input and output can have already be garbaged :if the   */
      /* socket is no more used, the input and output are not marked as   */
      /* used and can (eventually) be released before the call to         */
      /* shutdown (through free_socket) be done. One way could be to just */
      /* set SOCKET(sock).{in|out}put to #t and wait that next GC frees   */
      /* the ports if not already down. However, this will really         */
      /* disconnect the peer when the GC occurs rather than when the call */
      /* to shutdown is done. This is not important if this function is   */
      /* called by the GC, but could be annoying when it is called by the */
      /* user                                                             */
      if( INPUT_PORTP( SOCKET(sock).input ) ) {
	 bgl_close_input_port( SOCKET( sock ).input );
	 /* MS: 26 apr 2008, don't loose the port */
	 /* SOCKET( sock ).input = BFALSE; */
      }
   
      if( OUTPUT_PORTP( SOCKET( sock ).output ) ) {
	 bgl_close_output_port( SOCKET( sock ).output );
	 /* MS: 26 apr 2008, don't loose the port */
	 /* SOCKET(sock).output = BFALSE; */
      }

      return BUNSPEC;
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_socket_accept ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_socket_accept( obj_t serv, bool_t errp, obj_t inb, obj_t outb ) {
   char *ip;
   struct sockaddr_in sin;
   int len = sizeof( sin );
   int new_s;
   obj_t a_socket;
   int new_fd;
   obj_t hname;

   while( BAD_SOCKET( new_s = (int)accept( SOCKET( serv ).fd,
					   (struct sockaddr *)&sin,
					   (socklen_t *)&len)) ) {
      if( errno == EINTR )
	 continue;

      if( errp )
	 system_error( "socket-accept", serv );
      else
	 return BFALSE;
   }

   /* Tune the new socket */
#if( BGL_HAVE_SOCKET_TCP_CORK )
   setsockopt( new_s, IPPROTO_TCP, TCP_CORK, &sock_opt, sizeof( sock_opt ) );
#else   
#  if( BGL_HAVE_SOCKET_TCP_NODELAY )
   setsockopt( new_s, IPPROTO_TCP, TCP_NODELAY, &sock_opt, sizeof( sock_opt ) );
#  endif
#endif
   
   /* allocate the socket before any other allocation for bmem */
   a_socket = GC_MALLOC( SOCKET_SIZE );

   /* allocate and fill the new socket client for this connection */
   a_socket->socket_t.header = MAKE_HEADER( SOCKET_TYPE, 0 );
   a_socket->socket_t.portnum = ntohs( sin.sin_port );
   a_socket->socket_t.hostname = BUNSPEC;
   a_socket->socket_t.hostip = string_to_bstring( inet_ntoa( sin.sin_addr ) );
   a_socket->socket_t.fd = new_s;
   a_socket->socket_t.stype = SOCKET_CLIENT;
   a_socket->socket_t.userdata = BUNSPEC;

   set_socket_io_ports( new_s, BREF( a_socket ), "socket-accept", inb, outb );

   if( SOCKET( serv ).accept ) {
      /* this is used, for instance, by the ssl library */
      return SOCKET( serv ).accept( serv, BREF( a_socket ) );
   } else {
      return BREF( a_socket );
   }
}

/*---------------------------------------------------------------------*/
/*    long                                                             */
/*    bgl_socket_accept_many ...                                       */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF long
bgl_socket_accept_many( obj_t serv, bool_t errp, obj_t inbs, obj_t outbs, obj_t vec ) {
   int fd = SOCKET( serv ).fd;
   int flags;
   fd_set set;
   int n;
   long i;
   long l1 = VECTOR_LENGTH( inbs ), l2 = VECTOR_LENGTH( outbs );
   long l3 = VECTOR_LENGTH( vec );
   long l = l1 < l2 ? l1 : l2;


   if( l1 != l2 ) {
      C_SYSTEM_FAILURE( BGL_ERROR,
			"socket-accept-many",
			"in buffers and out buffers lengths mismatch",
			MAKE_PAIR( inbs, outbs ) );
   }
   
#if( BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   /* set the socket in non-blocking mode */
   /* this seems to be expensive so this should be done once for all */
   flags = fcntl( fd, F_GETFL );
   if( flags == -1 ) {
      if( errp )
	 system_error( "socket-accept-many", serv );
      else
	 return 0;
   }

   flags |= O_NONBLOCK;
   flags = fcntl( fd, F_SETFL, flags );

   if( flags == -1 ) {
      if( errp )
	 system_error( "socket-accept-many", serv );
      else
	 return 0;
   }

   /* select on the sock fd */
   FD_ZERO( &set );
   FD_SET( fd, &set );

   while( (n = select( fd + 1, &set, NULL, NULL, NULL )) <= 0 ) {

      if( errno == EINTR )
	 continue;

      if( errp )
	 C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			   "socket-accept-many",
			   strerror( errno ),
			   serv );
      else
	 return 0;
   }

   /* fill the result vector */
   for( i = 0; i < l; i++ ) {
      obj_t sock = bgl_socket_accept( serv,
				      0,
				      VECTOR_REF( inbs, i ),
				      VECTOR_REF( outbs, i ) );

      if( sock == BFALSE ) break;

      VECTOR_SET( vec, i, sock );
   }

   /* this seems to be expensive so this should be done once for all */
   /* WARNING: NOT TESTED... */
   flags &= ~O_NONBLOCK;
   flags = fcntl( fd, F_SETFL, flags );

   if( flags == -1 ) {
      system_error( "socket-accept-many", serv );
   }
   
   return i;
#else
   {
      obj_t sock = bgl_socket_accept( serv,
				      0,
				      VECTOR_REF( inbs, i ),
				      VECTOR_REF( outbs, i ) );
      VECTOR_SET( vec, 0, sock );

      return 1;
   }
#endif
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_socket_hostname ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_socket_hostname( obj_t sock ) {
   if( SOCKET( sock ).hostname == BUNSPEC ) {
      struct hostent *host;
#if( BGL_HAVE_INET_ATON )      
      struct sockaddr_in sin;

      inet_aton( BSTRING_TO_STRING( SOCKET( sock ).hostip ), &(sin.sin_addr) );
      
      if( host = bglhostbyaddr( &sin ) ) {
#else
      struct sockaddr_in *sin;

      sin = inet_addr( BSTRING_TO_STRING( SOCKET( sock ).hostip ) );
      
      if( host = bglhostbyaddr( sin ) ) {
#endif
	 return SOCKET( sock ).hostname = string_to_bstring( host->h_name );
      } else {
	 return SOCKET( sock ).hostname = SOCKET( sock ).hostip;
      }
   } else {
      return SOCKET( sock ).hostname;
   }      
}
      
/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    socket_close ...                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
socket_close( obj_t sock ) {
   int fd = SOCKET( sock ).fd;
   obj_t chook = SOCKET_CHOOK( sock );

   if( fd > 0 ) {
      /* MS: 19 Aug 2008, we don't have to close fd */
      /* since it will be closed automatically with */
      /* SOCKET( sock ).input                       */
      SOCKET( sock ).fd = -1;

      if( PROCEDUREP( chook ) ) {
	 if( PROCEDURE_ARITY( chook ) == 1 ) {
	    PROCEDURE_ENTRY( chook )( chook, sock, BEOA );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "socket-close",
			      "Illegal close hook arity",
			      chook );
	 }
      }

      if( INPUT_PORTP( SOCKET( sock ).input ) ) {
	 bgl_close_input_port( SOCKET( sock ).input );
	 /* MS: 26 apr 2008, don't loose the port */
	 /* SOCKET( sock ).input = BFALSE; */
      }
   
      if( OUTPUT_PORTP( SOCKET( sock ).output ) ) {
	 bgl_close_output_port( SOCKET( sock ).output );
	 /* MS: 26 apr 2008, don't loose the port */
	 /* SOCKET( sock ).output = BFALSE; */
      }
   }

   return BUNSPEC;
}
